"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import (
    List,
    Dict,
    DefaultDict,
    Set,
    Tuple,
    Union,
    Optional,
    Iterator,
    TextIO,
)

import os
from contextlib import contextmanager, suppress
from dataclasses import dataclass
from itertools import chain
from xml.etree.ElementTree import Element
import xml.etree.ElementTree as ET

from clang.cindex import TypeKind

from cparser_types import (
    CType,
    CPointerType,
    CRecordType,
    CFunctionType,
    CArrayType,
    CTypedefType,
    CPrimitiveType,
    assert_never,
)
from binding_class import Class, classes, class_structs
from binding_func import Func

from writer import Writer

doxygen_path: Optional[str] = None

DoxygenElements = Dict[str, Element]


class DoxygenFiles(Dict[str, DoxygenElements]):
    """
    Dictionary wrapper for retrieving XML elements from a filename

    If a filename is missing, parse and store its elements

    If an element name is duplicated, the element is ignored completely
    """

    names: Set[str]

    def __init__(self) -> None:
        self.names = set()
        super().__init__()

    def __missing__(self, path: str) -> DoxygenElements:
        tree = ET.parse(path)

        nodes: DoxygenElements = {}
        for node in tree.findall(
            "./compounddef/sectiondef/memberdef[@kind='function']"
        ):
            name = node.findtext("name")
            assert name
            if name in self.names:
                if name in nodes:
                    del nodes[name]
            else:
                self.names.add(name)
                nodes[name] = node
                super().__setitem__(path, nodes)

        return nodes


@dataclass
class DoxygenFile:
    """
    Groups a Doxygen XML file's name, refid, and absolute path
    """

    name: str
    refid: str
    path: str


doxygen_files = DoxygenFiles()


class DoxygenFunctions:
    """
    Wrapper for retrieving function XML element from a name
    """

    # Maps functions to filepaths
    function_files: DefaultDict[str, List[DoxygenFile]]

    def __init__(self) -> None:
        self.function_files = DefaultDict(list)

    def __getitem__(self, name: str) -> Element:
        files = self.function_files[name]

        if len(files) == 0:
            raise Exception(f"Function {name} not found in any file")

        if len(files) > 1:
            files = [f for f in files if not f.name.endswith(".h")]
            self.function_files[name] = files
            assert (
                len(files) == 1
            ), f"Function {name} is found in multiple files: {files}"

        return doxygen_files[files[0].path][name]


doxygen_functions = DoxygenFunctions()


def doxygen_children(node: Element) -> Iterator[Union[str, Element]]:
    """
    Create Iterator over text and node children of an Element
    """

    if node.text:
        yield node.text

    for child in node:
        yield child
        if child.tail:
            yield child.tail


class SphinxWriter(Writer):
    """
    Extended Writer class with additional sphinx-specific helpers
    """

    header_level: int

    def __init__(self, output: TextIO):
        super().__init__(output, indent_amount=3)
        self.header_level = 0

    def header(self, line: str, punctuation: str, *, overline: bool = False) -> None:
        """
        Write a sphinx header containing the specified text,
        and using the punctuation character as an underline
        (or optionally overline)
        """
        punctuation_line = punctuation * len(line)
        if overline:
            self.line(punctuation_line)
        self.line(line)
        self.line(punctuation_line)

    def title(self, title: str) -> None:
        """
        Write a sphinx title
        """
        self.header(title, "*", overline=True)

    @contextmanager
    def directive(
        self, name: str, *args: str, options: Optional[List[Tuple[str, str]]] = None
    ) -> Iterator[None]:
        """
        Write a sphinx directive, with *args as arguments
        and options as options

        Also increases indentation for duration of context
        """
        if not args:
            self.line(f".. {name}::")
        else:
            self.line(f".. {name}:: {args[0]}")
            spacer = " " * (5 + len(name))
            for arg in args[1:]:
                self.line(f"{spacer} {arg}")

        self.indent_level += 1

        if options:
            for option_name, value in options:
                self.line(f":{option_name}: {value}")

        self.line("")
        yield

        self.indent_level -= 1
        self.line("")


def generate(output_dir: str) -> None:
    """
    Generate sphinx docs and write to sphinx/ directory
    """
    sphinx_dir = os.path.join(output_dir, "sphinx")

    with suppress(FileExistsError):
        os.mkdir(sphinx_dir)

    with open(os.path.join(sphinx_dir, "conf.py"), "w", encoding="utf-8") as output:
        writer = Writer(output)
        writer.line(
            "html_theme = 'furo'",
            "",
            "import shutil",
            "import os",
            "def setup(app):",
            "    shutil.copytree(",
            f"        os.path.join('{doxygen_path}', 'html'),",
            "        os.path.join(app.outdir, 'doxygen'),",
            "        dirs_exist_ok=True",
            "    )",
        )

    with open(os.path.join(sphinx_dir, "index.rst"), "w", encoding="utf-8") as output:
        writer = SphinxWriter(output)
        writer.title("Rizin Python Bindings")
        writer.line(
            ".. toctree::",
            "   classes",
        )

    if doxygen_path:
        tree = ET.parse(os.path.join(doxygen_path, "xml", "index.xml"))

        for filenode in tree.findall("./compound[@kind='file']"):
            refid = filenode.attrib["refid"]
            filename = filenode.findtext("name")
            assert filename

            for membernode in filenode.findall("./member"):
                name = membernode.findtext("name")
                kind = membernode.attrib["kind"]
                if kind == "function":
                    assert name
                    doxygen_functions.function_files[name].append(
                        DoxygenFile(
                            name=filename,
                            refid=refid,
                            path=os.path.join(doxygen_path, "xml", f"{refid}.xml"),
                        )
                    )

    generate_classes(sphinx_dir)


def generate_classes(sphinx_dir: str) -> None:
    """
    Generate sphinx classes docs

    Writes documentation for each class
    in sphinx_output/classes directory

    Writes each class doc file to the table of contents
    tree in classes.rst
    """
    with open(os.path.join(sphinx_dir, "classes.rst"), "w", encoding="utf-8") as output:
        writer = SphinxWriter(output)
        writer.title("Classes")

        with writer.directive("toctree"):
            for classname in sorted(classes):
                writer.line(f"classes/{classname}")

    classes_dir = os.path.join(sphinx_dir, "classes")
    with suppress(FileExistsError):
        os.mkdir(classes_dir)

    for name, cls in classes.items():
        with open(
            os.path.join(classes_dir, f"{name}.rst"), "w", encoding="utf-8"
        ) as output:
            write_class(SphinxWriter(output), cls)


def write_class(writer: SphinxWriter, cls: Class) -> None:
    """
    Generate sphinx class
    """
    writer.title(cls.name)

    with writer.directive("py:class", cls.name):
        for name, field in sorted(cls.fields.items()):
            with writer.directive(
                "py:property",
                field.rename or name,
                options=[("type", stringify_ctype(field.ctype))],
            ):
                pass

        def stringify_func(name: str, func: Func, method: bool = True) -> str:
            args = []
            for arg in func.cfunc.args[1:] if method else func.cfunc.args:
                arg_name = arg.cursor.spelling
                if arg_name == "self":
                    arg_name = "_self"
                args.append(f"{arg_name}: {stringify_ctype(arg.ctype)}")

            args_str = ", ".join(args)
            result = stringify_ctype(func.cfunc.result_ctype)
            return f"{name}({args_str}) -> {result}"

        for directive, func_dict in [
            ("py:staticmethod", cls.funcs),
            ("py:method", cls.methods),
        ]:
            for name, func in sorted(func_dict.items()):
                with writer.directive(directive, stringify_func(name, func)):
                    if "RZ_DEPRECATE" in func.cfunc.attrs:
                        with writer.directive("warning"):
                            writer.line(
                                f"Calls deprecated function ``{func.cfunc.cursor.spelling}``"
                            )

                    if doxygen_path:
                        write_doxygen_function(writer, func.cfunc.cursor.spelling)


doxygen_escape_table = str.maketrans({"*": "\\*", "_": "\\_"})


def stringify_doxygen_computeroutput(node: Element) -> str:
    """
    Parse doxygen <computeroutput> tag and
    convert to reST ``internal literal``
    """

    segments = ["``"]

    for child in doxygen_children(node):
        if isinstance(child, str):
            segments.append(child)
        elif child.tag == "ref":
            assert child.text
            segments.append(child.text)
        else:
            raise Exception(
                f"Unknown doxygen tag in child of computeroutput: {child.tag}"
            )

    segments.append("``")
    return "".join(segments)


def write_doxygen_parameterlist(writer: SphinxWriter, node: Element) -> None:
    """
    Generate sphinx :param: options from doxygen <parameterlist> element
    """

    kind = node.attrib["kind"]

    if kind == "param":
        for item in node.findall("./parameteritem"):
            names = item.findall("./parameternamelist/parametername")
            assert len(names) == 1
            name = names[0].text

            if name == "self":
                name = "_self"

            segments = []

            paras = item.findall("./parameterdescription/para")
            assert len(paras) == 1
            para = paras[0]

            for child in doxygen_children(para):
                if isinstance(child, str):
                    segments.append(child.translate(doxygen_escape_table))
                elif child.tag == "ref":
                    assert child.text
                    segments.append(child.text)
                elif child.tag == "computeroutput":
                    segments.append(stringify_doxygen_computeroutput(child))
                else:
                    raise Exception(
                        f"Unknown doxygen tag in child of parameterdescription/para: {child.tag}"
                    )

            contents = "".join(segments)
            writer.line(f":param {name}: {contents}")
    else:
        raise Exception(f"Unknown doxygen parameteritem kind {kind}")


def write_doxygen_simplesect(writer: SphinxWriter, node: Element) -> None:
    """
    Generate sphinx :return: options from doxygen <simplesect> element
    """

    kind = node.attrib["kind"]

    if kind == "return":
        segments = []

        paras = node.findall("para")
        assert len(paras) == 1
        para = paras[0]

        for child in doxygen_children(para):
            if isinstance(child, str):
                segments.append(child.translate(doxygen_escape_table))
            elif child.tag == "ref":
                assert child.text
                segments.append(child.text)
            elif child.tag == "computeroutput":
                segments.append(stringify_doxygen_computeroutput(child))
            else:
                raise Exception(
                    f"Unknown doxygen tag in child of simpleselect/para: {child.tag}"
                )

        contents = "".join(segments)
        writer.line(f":return: {contents}", "")

    elif kind == "see":
        pass
    else:
        raise Exception(f"Unknown doxygen simplesect kind {kind}")


def write_doxygen_function(writer: SphinxWriter, name: str) -> None:
    """
    Generate documentation for function from doxygen xml
    """
    element = doxygen_functions[name]

    doxygen_file = doxygen_functions.function_files[name][0]
    url_hash = element.attrib["id"][-33:]
    writer.line(
        f"Calls function ``{name}``",
        f"(defined in `{doxygen_file.name} <../doxygen/{doxygen_file.refid}.html#{url_hash}>`__)",
        "",
    )

    for paragraph in chain.from_iterable(
        element.findall(xpath)
        for xpath in ["./briefdescription/para", "./detaileddescription/para"]
    ):
        segments = []

        for node in doxygen_children(paragraph):
            if isinstance(node, str):
                segments.append(node.translate(doxygen_escape_table))
            elif node.tag == "computeroutput":
                segments.append(stringify_doxygen_computeroutput(node))

            elif node.tag in "itemizedlist":
                writer.line("".join(segments), "")  # Flush segments

                for item in node.findall("./listitem/para"):
                    segments = []
                    for child in doxygen_children(item):
                        assert isinstance(child, str)
                        segments.append(child.translate(doxygen_escape_table))
                    contents = "".join(segments)
                    writer.line(f"- {contents}")

                segments = []
            elif node.tag == "parameterlist":
                writer.line("".join(segments), "")  # Flush segments
                segments = []

                write_doxygen_parameterlist(writer, node)

            elif node.tag == "ref":
                assert node.text
                segments.append(node.text)

            elif node.tag == "simplesect":
                writer.line("".join(segments), "")  # Flush segments
                segments = []

                write_doxygen_simplesect(writer, node)
            elif node.tag == "ulink":
                url = node.attrib["url"]
                text = node.text
                assert text
                segments.append(f"`{text} <{url}>`__")
            else:
                raise Exception(f"Unknown doxygen tag in child of para: {node.tag}")

        writer.line("".join(segments), "")


def stringify_ctype(ctype: CType) -> str:
    """
    Translate a CType to a Python type annotation
    """

    inner = None
    levels = 0

    while True:
        if isinstance(ctype, CPointerType):
            ctype = ctype.pointee
        elif isinstance(ctype, CArrayType):
            ctype = ctype.element
        else:
            break

        kind = ctype.type_.kind
        if levels == 0:
            if kind == TypeKind.VOID:
                inner = "Any"
                break
            if kind in [TypeKind.SCHAR, TypeKind.CHAR_S]:
                inner = "str"
                break

        levels += 1

    def stringify_ctype_primitive(ctype: CType) -> str:
        if isinstance(ctype, CPrimitiveType):
            kind = ctype.type_.kind
            if kind == TypeKind.VOID:
                return "None"

            if kind == TypeKind.BOOL:
                return "bool"

            if kind in [TypeKind.FLOAT, TypeKind.DOUBLE]:
                return "float"

            return "int"

        if isinstance(ctype, CRecordType):
            if ctype.generic:
                return f"{ctype.generic.name}[{ctype.specialization}]"
            return class_structs[ctype.decl_spelling].name

        if isinstance(ctype, CTypedefType):
            if isinstance(ctype.canonical, CRecordType) and ctype.canonical.generic:
                return stringify_ctype(ctype.canonical)
            return ctype.cursor.spelling

        if isinstance(ctype, CFunctionType):
            result = stringify_ctype(ctype.result)
            args = ", ".join(stringify_ctype(arg) for arg in ctype.args)
            return f"CFunction[[{args}], {result}]"

        assert not isinstance(ctype, CPointerType) and not isinstance(ctype, CArrayType)
        assert_never(ctype)

    if not inner:
        inner = stringify_ctype_primitive(ctype)

    if levels == 1:
        levels = 0

    return ("Pointer[" * levels) + inner + ("]" * levels)



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import OrderedDict, Callable

import concurrent.futures

from cparser_header import HeaderBuilder, Header
from binding_class import Class
from binding_director import Director
from binding_enum import Enum, MacroEnum
from binding_generic import Generic

HeaderFunc = Callable[[Header], None]
threaded_headers: OrderedDict[str, HeaderFunc] = OrderedDict()


def threaded_header(name: str) -> Callable[[HeaderFunc], None]:
    """
    Registers a header to be parsed by clang in parallel

    Decorates a function which takes a Header as an argument.

    Only the header parsing is done in parallel; the wrapped
    functions are called in the order they were registered
    """

    def decorator(func: HeaderFunc) -> None:
        assert name not in threaded_headers
        threaded_headers[name] = func

    return decorator


def run() -> None:
    """
    Parse headers in parallel, then run registered functions sequentially
    """

    with concurrent.futures.ThreadPoolExecutor() as executor:
        builders = [HeaderBuilder(name) for name in threaded_headers]
        translation_units = executor.map(HeaderBuilder.translation_unit, builders)
        for translation_unit, builder, func in zip(
            translation_units, builders, threaded_headers.values()
        ):
            for diagnostic in translation_unit.diagnostics:
                print(diagnostic)

            func(Header(translation_unit, builder))


############
# GENERICS #
############


@threaded_header("rz_list.h")
def bind_list(list_h: Header) -> None:
    """
    RzListIter, RzList
    """
    ### RzListIter ###
    rz_list_iter = Generic(list_h, "RzListIter", pointer=True)
    rz_list_iter.add_method("rz_list_iter_get_next", rename="next", generic_ret=True)
    rz_list_iter.add_method("rz_list_iter_get_data", rename="data", generic_ret=True)

    ### RzList ###
    rz_list = Generic(list_h, "RzList", dependencies=[rz_list_iter], pointer=True)
    rz_list.add_method("rz_list_length", rename="length")

    rz_list.add_method("rz_list_first", rename="first", generic_ret=True)
    rz_list.add_method("rz_list_last", rename="last", generic_ret=True)
    rz_list.add_method("rz_list_iterator", rename="iterator", generic_ret=True)

    rz_list.add_method(
        "rz_list_prepend", rename="prepend", generic_ret=True, generic_args={"data"}
    )
    rz_list.add_method(
        "rz_list_append", rename="append", generic_ret=True, generic_args={"data"}
    )

    rz_list.add_python_method("__len__(self)", "return self.length()")
    rz_list.add_python_method("__iter__(self)", "return RzListIterator(self)")

    # Specialized constructors
    rz_list.add_specialization_extension(
        "RzBinSymbol",
        "RzList_RzBinSymbol() {",
        "    return rz_list_newf((RzListFree)rz_bin_symbol_free);",
        "}",
    )
    rz_list.add_specialization_extension(
        "RzBinSection",
        "RzList_RzBinSection() {",
        "    return rz_list_newf((RzListFree)rz_bin_section_free);",
        "}",
    )
    rz_list.add_specialization_extension(
        "RzBinMap",
        "RzList_RzBinMap() {",
        "    return rz_list_newf((RzListFree)rz_bin_map_free);",
        "}",
    )


@threaded_header("rz_vector.h")
def bind_vector(vector_h: Header) -> None:
    """
    RzVector, RzPVector
    """
    ### RzVector ###
    rz_vector = Generic(vector_h, "RzVector")
    rz_vector.add_method("rz_vector_len", rename="length")
    rz_vector.add_method("rz_vector_head", rename="head", generic_ret=True)
    rz_vector.add_method("rz_vector_tail", rename="tail", generic_ret=True)
    rz_vector.add_method("rz_vector_index_ptr", rename="at", generic_ret=True)
    rz_vector.add_method(
        "rz_vector_push", rename="push", generic_ret=True, generic_args={"x"}
    )

    rz_vector.add_python_method("__len__(self)", "return self.length()")
    rz_vector.add_python_method("__iter__(self)", "return RzVectorIterator(self)")

    ### RzPVector ###
    rz_pvector = Generic(vector_h, "RzPVector", pointer=True)
    rz_pvector.add_method("rz_pvector_len", rename="length")
    rz_pvector.add_method("rz_pvector_head", rename="head", generic_ret=True)
    rz_pvector.add_method("rz_pvector_tail", rename="tail", generic_ret=True)
    rz_pvector.add_method("rz_pvector_at", rename="at", generic_ret=True)
    rz_pvector.add_method(
        "rz_pvector_push", rename="push", generic_ret=True, generic_args={"x"}
    )

    rz_pvector.add_python_method("__len__(self)", "return self.length()")
    rz_pvector.add_python_method("__iter__(self)", "return RzPVectorIterator(self)")


###########
# HEADERS #
###########


@threaded_header("rz_analysis.h")
def bind_analysis(analysis_h: Header) -> None:
    """
    RzAnalysis
    """
    analysis_h.ignore("rz_analysis_version")

    rz_analysis = Class(
        analysis_h,
        typedef="RzAnalysis",
        ignore_fields={"leaddrs"},
        rename_fields={},
    )

    rz_analysis_function = Class(analysis_h, typedef="RzAnalysisFunction")
    rz_analysis_function.add_method("rz_analysis_function_delete", rename="delete_self")
    rz_analysis_function.add_prefixed_methods("rz_analysis_function_")

    rz_analysis.add_method("rz_analysis_reflines_get", rename="get_reflines")
    rz_analysis.add_prefixed_methods("rz_analysis_")
    rz_analysis.add_prefixed_funcs("rz_analysis_")

    Class(analysis_h, typedef="RzAnalysisBlock")
    Class(analysis_h, typedef="RzAnalysisEsil")
    Class(analysis_h, typedef="RzAnalysisPlugin")

    Class(
        analysis_h,
        typedef="RzAnalysisXRef",
        rename_fields={"from": "from_addr", "to": "to_addr"},
    )
    Enum(analysis_h, typedef="RzAnalysisXRefType")


@threaded_header("rz_asm.h")
def bind_asm(asm_h: Header) -> None:
    """
    RzAsm
    """
    Class(asm_h, typedef="RzAsm")  # TODO: Add functions
    Class(asm_h, typedef="RzAsmPlugin")


@threaded_header("rz_bin.h")
def bind_bin(bin_h: Header) -> None:
    """
    RzBin
    """
    bin_h.ignore("rz_bin_version")

    rz_bin_file = Class(bin_h, typedef="RzBinFile")
    rz_bin_file.add_prefixed_methods("rz_bin_file_")

    rz_bin = Class(
        bin_h, typedef="RzBin", rename_fields={"cur": "_cur", "strpurge": "_strpurge"}
    )
    rz_bin.add_prefixed_methods("rz_bin_")
    rz_bin.add_prefixed_funcs("rz_bin_")

    Class(bin_h, typedef="RzBinXtrPlugin")
    Class(bin_h, typedef="RzBinOptions")
    Class(bin_h, typedef="RzBinInfo")
    Class(bin_h, typedef="RzBinSymbol")
    Class(bin_h, typedef="RzBinSection")
    Class(bin_h, typedef="RzBinMap")

    Director(bin_h, "RzBinPlugin")

    MacroEnum(bin_h, prefix="RZ_BIN_TYPE_")
    MacroEnum(bin_h, prefix="RZ_BIN_BIND_")


@threaded_header("rz_util/rz_buf.h")
def bind_buf(buf_h: Header) -> None:
    """
    RzBuf
    """
    rz_buf = Class(buf_h, typedef="RzBuffer")
    rz_buf.add_prefixed_methods("rz_buf_")

    MacroEnum(buf_h, "RZ_BUF_SET", "RZ_BUF_CUR", "RZ_BUF_END")


@threaded_header("rz_cmd.h")
def bind_cmd(cmd_h: Header) -> None:
    """
    RzCmd
    """
    Class(cmd_h, typedef="RzCmd")
    Class(cmd_h, typedef="RzCmdDescHelp")
    Class(cmd_h, typedef="RzCmdDescArg")
    Enum(cmd_h, typedef="RzCmdArgType")


@threaded_header("rz_config.h")
def bind_config(config_h: Header) -> None:
    """
    RzConfig
    """
    config_h.ignore("rz_config_version")  # TODO: Fix in rizin
    rz_config = Class(config_h, typedef="RzConfig", rename_fields={"lock": "_lock"})
    rz_config.add_prefixed_methods("rz_config_")
    rz_config.add_prefixed_funcs("rz_config_")


@threaded_header("rz_cons.h")
def bind_cons(cons_h: Header) -> None:
    """
    RzCons
    """
    cons_h.ignore("rz_cons_version")

    rz_cons = Class(
        cons_h,
        typedef="RzCons",
        ignore_fields={"term_raw", "term_buf"},  # struct termios
        rename_fields={
            "lastline": "_lastline",
            "echo": "_echo",
            "highlight": "_highlight",
            "newline": "_newline",
            "filter": "_filter",
            "flush": "_flush",
            "input": "_input",
            "enable_highlight": "_enable_highlight",
        },
    )

    cons_h.ignore(
        "rz_cons_printf",
        "rz_cons_printf_list",
        "rz_cons_yesno",
    )
    rz_cons.add_prefixed_methods("rz_cons_")
    rz_cons.add_prefixed_funcs("rz_cons_")

    Class(cons_h, typedef="RzLine")


@threaded_header("rz_cmp.h")
def bind_cmp(_: Header) -> None:
    """
    Needed for RzList_RzCoreCmpWatcher
    """


@threaded_header("rz_core.h")
def bind_core(core_h: Header) -> None:
    """
    RzCore
    """
    core_h.ignore("rz_core_version")

    rz_core = Class(
        core_h,
        typedef="RzCore",
        struct="rz_core_t",
        rename_fields={"autocomplete": "_autocomplete", "visual": "_visual"},
    )
    rz_core.add_constructor("rz_core_new")
    rz_core.add_destructor("rz_core_free")

    # Ignore format string functions
    core_h.ignore(
        "rz_core_notify_begin",
        "rz_core_notify_done",
        "rz_core_notify_error",
        "rz_core_cmd_strf",
        "rz_core_cmdf",
        "rz_core_syscallf",
    )

    rz_core.add_method(
        "rz_core_file_open_load",
        rename="file_open_load",
        default_args={"addr": "0", "perms": "0", "write_mode": "0"},
    )
    rz_core.add_method(
        "rz_core_file_open",
        rename="file_open",
        default_args={"flags": "0", "loadaddr": "UT64_MAX"},
    )
    rz_core.add_prefixed_methods("rz_core_")
    rz_core.add_prefixed_funcs("rz_core_")

    Class(core_h, typedef="RzCoreFile")


@threaded_header("rz_flag.h")
def bind_flag(flag_h: Header) -> None:
    """
    RzFlag
    """
    flag_h.ignore("rz_flag_version")

    rz_flag_item = Class(flag_h, typedef="RzFlagItem")
    rz_flag_item.add_prefixed_methods("rz_flag_item_")

    rz_flag = Class(flag_h, typedef="RzFlag", ignore_fields={"tags"})
    rz_flag.add_prefixed_methods("rz_flag_")
    rz_flag.add_prefixed_funcs("rz_flag_")


@threaded_header("rz_hash.h")
def bind_hash(hash_h: Header) -> None:
    """
    RzHash
    """
    Class(hash_h, typedef="RzHash")  # TODO: Add functions


@threaded_header("rz_io.h")
def bind_io(io_h: Header) -> None:
    """
    RzIO
    """
    Class(io_h, typedef="RzIO", ignore_fields={"ptrace_wrap", "priv_w32dbg_wrap"})
    Class(io_h, typedef="RzIODesc")
    Class(io_h, typedef="RzIOPlugin")


@threaded_header("rz_util/rz_num.h")
def bind_num(num_h: Header) -> None:
    """
    RzNum
    """
    rz_num = Class(num_h, typedef="RzNum")
    rz_num.add_prefixed_methods("rz_num_")
    rz_num.add_prefixed_funcs("rz_num_")


@threaded_header("rz_reg.h")
def bind_reg(reg_h: Header) -> None:
    """
    RzReg
    """
    reg_h.ignore("rz_reg_version")

    rz_reg = Class(reg_h, typedef="RzReg", rename_fields={"regset": "_regset"})
    rz_reg.add_method("rz_reg_32_to_64", rename="convert_32_to_64")
    rz_reg.add_method("rz_reg_64_to_32", rename="convert_64_to_32")
    rz_reg.add_prefixed_methods("rz_reg_")
    rz_reg.add_prefixed_funcs("rz_reg_")


@threaded_header("rz_type.h")
def bind_type(type_h: Header) -> None:
    """
    RzType
    """
    type_h.ignore("rz_type_version")
    rz_type = Class(type_h, typedef="RzType", rename_fields={"kind": "_kind"})
    rz_type.add_prefixed_methods("rz_type_")
    rz_type.add_prefixed_funcs("rz_type_")


@threaded_header("rz_types.h")
def bind_types(types_h: Header) -> None:
    """
    RZ_PERM_R/W/X
    """
    MacroEnum(types_h, prefix="RZ_PERM_")


#######
# SDB #
#######
@threaded_header("sdb/sdb.h")
def bind_sdb(sdb_h: Header) -> None:
    """
    sdb
    """
    Class(sdb_h, typedef="Sdb", ignore_fields={"db", "m"})


@threaded_header("sdb/ls.h")
def bind_ls(ls_h: Header) -> None:
    """
    ls
    """
    Class(ls_h, typedef="SdbList")


@threaded_header("sdb/ht_pp.h")
def bind_ht_pp(ht_pp_h: Header) -> None:
    """
    ht_pp
    """
    Class(ht_pp_h, typedef="HtPP")


@threaded_header("sdb/ht_pu.h")
def bind_ht_pu(ht_pu_h: Header) -> None:
    """
    ht_pu
    """
    Class(ht_pu_h, typedef="HtPU")


@threaded_header("sdb/ht_up.h")
def bind_ht_up(ht_up_h: Header) -> None:
    """
    ht_up
    """
    Class(ht_up_h, typedef="HtUP")


@threaded_header("sdb/ht_uu.h")
def bind_ht_uu(ht_uu_h: Header) -> None:
    """
    ht_uu
    """
    Class(ht_uu_h, typedef="HtUU")



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import List, Optional, Union, NoReturn, TYPE_CHECKING

from dataclasses import dataclass

from clang.cindex import Cursor, Type, TypeKind

if TYPE_CHECKING:
    from binding_generic import Generic


@dataclass
class CBaseType:
    """
    Base class for type wrapper
    """

    type_: Type


@dataclass
class CPrimitiveType(CBaseType):
    """
    Primitive type wrapper
    """


@dataclass
class CPointerType(CBaseType):
    """
    Pointer type wrapper
    """

    pointee: "CType"


@dataclass
class CArrayType(CBaseType):
    """
    Array type wrapper

    If fixed size (eg. char[80]), element_count is an int.
    If variable size (eg. char[]), element_count is None.
    """

    element: "CType"
    element_count: Optional[int] = None


@dataclass
class CTypedefType(CBaseType):
    """
    Typedef type wrapper
    """

    canonical: "CType"
    cursor: Cursor


@dataclass
class CRecordType(CBaseType):
    """
    Struct type wrapper

    Structs without a name (eg. typedef struct {} AStruct)
    use their typedef name for decl_spelling
    """

    decl_spelling: str
    generic: Optional["Generic"] = None
    specialization: Optional[str] = None


@dataclass
class CFunctionType(CBaseType):
    """
    Function type wrapper
    """

    result: "CType"
    args: List["CType"]
    arg_names: Optional[List[str]] = None


CType = Union[
    CPrimitiveType,
    CRecordType,
    CPointerType,
    CFunctionType,
    CArrayType,
    CTypedefType,
]


def wrap_type(type_: Type) -> CType:
    """
    Wrap a type
    """
    while type_.kind == TypeKind.ELABORATED:
        type_ = type_.get_named_type()

    # Complex types
    if type_.kind == TypeKind.POINTER:
        return CPointerType(type_, pointee=wrap_type(type_.get_pointee()))

    if type_.kind in [TypeKind.CONSTANTARRAY, TypeKind.INCOMPLETEARRAY]:
        array_t = CArrayType(type_, element=wrap_type(type_.get_array_element_type()))
        if type_.kind == TypeKind.CONSTANTARRAY:
            array_t.element_count = type_.element_count
        return array_t

    if type_.kind == TypeKind.TYPEDEF:
        return CTypedefType(
            type_,
            canonical=wrap_type(type_.get_canonical()),
            cursor=type_.get_declaration(),
        )

    if type_.kind == TypeKind.FUNCTIONPROTO:
        return CFunctionType(
            type_,
            result=wrap_type(type_.get_result()),
            args=[wrap_type(arg) for arg in type_.argument_types()],
        )

    if type_.kind == TypeKind.RECORD:
        decl_spelling = type_.get_declaration().spelling or type_.spelling
        if decl_spelling.startswith("const "):
            decl_spelling = decl_spelling[len("const ") :]

        return CRecordType(type_, decl_spelling=decl_spelling)

    # Primitive types
    if type_.kind in [
        TypeKind.ENUM,
        TypeKind.VOID,
        TypeKind.BOOL,
        TypeKind.FLOAT,
        TypeKind.DOUBLE,
        TypeKind.LONGDOUBLE,
        # Unsigned
        TypeKind.CHAR_U,
        TypeKind.UCHAR,
        TypeKind.CHAR16,
        TypeKind.CHAR32,
        TypeKind.USHORT,
        TypeKind.UINT,
        TypeKind.ULONG,
        TypeKind.ULONGLONG,
        # Signed
        TypeKind.CHAR_S,
        TypeKind.SCHAR,
        TypeKind.WCHAR,
        TypeKind.SHORT,
        TypeKind.INT,
        TypeKind.LONG,
        TypeKind.LONGLONG,
    ]:
        return CPrimitiveType(type_)
    raise Exception(f"Unknown type kind {type_.kind}")


def assert_never(_: NoReturn) -> NoReturn:
    """
    Helper to typecheck exhaustiveness
    """
    assert False



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import List, Dict, OrderedDict, Set, Optional, overload

from dataclasses import dataclass

from clang.cindex import CursorKind

from cparser_header import Header
from cparser_types import CType, CPointerType, CTypedefType, CRecordType, wrap_type
from binding_func import Func
from binding_generic_specializations import gen_ctype_specializations
from binding_typemap import Typemap

classes: OrderedDict[str, "Class"] = OrderedDict()
class_structs: Dict[str, "Class"] = {}


@dataclass
class Field:
    """
    Groups a field name and type
    """

    name: str
    rename: Optional[str]
    ctype: CType


class Class:
    """
    A collection of fields, static functions, and methods wrapping a C struct
    """

    name: str
    struct_name: str
    header: Header

    fields: OrderedDict[str, Field]
    funcs: OrderedDict[str, Func]
    methods: OrderedDict[str, Func]

    constructor: Optional[Func]
    destructor: Optional[Func]

    @overload
    def __init__(
        self,
        header: Header,
        *,
        typedef: str,
        ignore_fields: Optional[Set[str]] = ...,
        rename_fields: Optional[Dict[str, str]] = ...,
    ):
        ...

    @overload
    def __init__(
        self,
        header: Header,
        *,
        typedef: str,
        struct: str,
        ignore_fields: Optional[Set[str]] = ...,
        rename_fields: Optional[Dict[str, str]] = ...,
    ):
        ...

    def __init__(
        self,
        header: Header,
        *,
        typedef: str,
        struct: Optional[str] = None,
        ignore_fields: Optional[Set[str]] = None,
        rename_fields: Optional[Dict[str, str]] = None,
    ):
        self.header = header
        self.name = typedef

        self.fields = OrderedDict()
        self.funcs = OrderedDict()
        self.methods = OrderedDict()
        self.constructor = None
        self.destructor = None

        # Get struct cursor from header
        if not struct:
            typedef_cursor = header.pop(CursorKind.TYPEDEF_DECL, typedef)
            struct_cursor = typedef_cursor.underlying_typedef_type.get_declaration()
            assert (
                struct_cursor.kind == CursorKind.STRUCT_DECL
            ), "Typedef underlying declaration was {struct_cursor.kind}, not STRUCT_DECL"
        else:
            struct_cursor = header.pop(CursorKind.STRUCT_DECL, struct)

        self.struct_name = struct_cursor.spelling or typedef

        assert typedef not in classes
        classes[typedef] = self
        class_structs[self.struct_name] = self

        # Parse struct fields
        for field in struct_cursor.get_children():
            if field.kind == CursorKind.FIELD_DECL:
                name = field.spelling
                if ignore_fields and name in ignore_fields:
                    ignore_fields.remove(name)
                    continue

                if rename_fields:
                    rename = rename_fields.pop(name, None)
                else:
                    rename = None

                assert name not in self.fields
                ctype = wrap_type(field.type)
                gen_ctype_specializations([field], ctype)
                self.fields[name] = Field(name, rename, ctype)

            elif field.kind not in [CursorKind.STRUCT_DECL, CursorKind.UNION_DECL]:
                raise Exception(
                    f"Unexpected struct child of kind: {field.kind} at {field.location}"
                )

        # Ensure all ignore_fields and rename_fields are for valid fields
        if ignore_fields and len(ignore_fields) != 0:
            print(
                f"[WARNING] Ignored fields on class {self.name} do not exist:",
                ", ".join(ignore_fields),
            )
        if rename_fields and len(rename_fields) != 0:
            print(
                f"[WARNING] Renamed fields on class {self.name} do not exist:",
                ", ".join(rename_fields),
            )

    def add_constructor(self, name: str) -> None:
        """
        Set C function with name as class constructor
        """
        assert not self.constructor
        self.constructor = Func(self.header, name)

    def add_destructor(self, name: str) -> None:
        """
        Set C function with name as class destructor
        """
        assert not self.destructor
        self.destructor = Func(self.header, name)

    def add_func(
        self,
        name: str,
        *,
        rename: str,
        default_args: Optional[Dict[str, str]] = None,
        typemaps: Optional[List[Typemap]] = None,
    ) -> None:
        """
        Add C function with name as static function
        """
        func = Func(
            self.header,
            name,
            default_args=default_args,
            typemaps=typemaps,
        )

        assert rename not in self.funcs
        self.funcs[rename] = func

    def add_method(
        self,
        name: str,
        *,
        rename: str,
        default_args: Optional[Dict[str, str]] = None,
        typemaps: Optional[List[Typemap]] = None,
    ) -> None:
        """
        Add C function with name as method
        """
        method = Func(
            self.header,
            name,
            default_args=default_args,
            typemaps=typemaps,
        )

        assert rename not in self.methods
        self.methods[rename] = method

    def add_prefixed_methods(self, prefix: str) -> None:
        """
        Add C functions with a given prefix, and that take a
        pointer to this class as the first argument, as methods

        New names will be the function names with the prefix removed

        Note that this may result in names beginning with numbers
        (eg. rz_reg_64_to_32 -> 64_to_32) which need to be manually added
        """
        method_names = set()
        for name, cfunc in self.header.cfuncs.items():
            if not name.startswith(prefix):
                continue
            if "RZ_API" not in cfunc.attrs:
                continue
            if len(cfunc.args) == 0:
                continue

            # Select functions with class struct as first argument
            ctype = cfunc.args[0].ctype
            if not isinstance(ctype, CPointerType):
                continue

            ctype = ctype.pointee
            if isinstance(ctype, CTypedefType):
                ctype = ctype.canonical

            if not isinstance(ctype, CRecordType):
                continue

            if ctype.decl_spelling != self.struct_name:
                continue

            rename = name[len(prefix) :]
            method = Func(
                self.header,
                cfunc=cfunc,
            )

            method_names.add(name)
            assert rename not in self.methods
            self.methods[rename] = method

        self.header.ignore(*method_names)

    def add_prefixed_funcs(self, prefix: str) -> None:
        """
        Add C functions with a given prefix as static functions
        """
        func_names = set()
        for name, cfunc in self.header.cfuncs.items():
            if not name.startswith(prefix):
                continue
            if "RZ_API" not in cfunc.attrs:
                continue

            rename = name[len(prefix) :]
            func = Func(
                self.header,
                cfunc=cfunc,
            )

            func_names.add(name)
            assert rename not in self.funcs
            self.funcs[rename] = func

        self.header.ignore(*func_names)



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

import os
from enum import Enum as PyEnum

from clang.cindex import TypeKind

from cparser_header import headers
from cparser_types import (
    CType,
    CPointerType,
    CRecordType,
    CFunctionType,
    CArrayType,
    CTypedefType,
    CPrimitiveType,
    assert_never,
)
from binding_generic import Generic, generics
from binding_generic_specializations import generic_structs
from binding_class import Class, classes, class_structs
from binding_func import Func, GenericFunc
from binding_director import Director, directors
from binding_enum import Enum, enums, macro_enums
from writer import Writer


def generate(output_dir: str) -> None:
    """
    Generate SWIG bindings and write to output_dir/rizin.i
    """
    output_path = os.path.join(output_dir, "rizin.i")
    with open(output_path, "w", encoding="utf-8") as output_file:
        write(Writer(output_file))


def write(writer: Writer) -> None:
    """
    Toplevel SWIG generator
    """
    writer.line("%module(directors=1) rizin")
    writer.line("%{")
    for header in headers:
        writer.line(f"#include <{header.name}>")
    writer.line("%}")

    writer.snippet("snippets_swig/prologue.i")
    writer.snippet("snippets_swig/cmd_director.i")

    writer.line("%pythoncode %{")
    writer.snippet("snippets_swig/iterators.py")
    writer.line("%}")

    for generic in generics.values():
        write_generic(writer, generic)

    for cls in classes.values():
        write_class(writer, cls)

    for director in directors.values():
        write_director(writer, director)

    for enum in enums:
        write_enum(writer, enum)

    for macro_enum in macro_enums:
        for name, definition in macro_enum.defines.items():
            writer.line(f"#define {name} {definition}")

    writer.line("%extend rz_cmd_t {")
    writer.snippet("snippets_swig/register_swig_command.cpp")  # TODO: %catches
    writer.line("}")

    writer.line("%extend rz_core_t {", "%pythoncode %{")
    writer.snippet("snippets_swig/register_command.py")
    writer.line("%}", "}")


class FuncKind(PyEnum):
    """
    Python enumeration for kind of class function
    """

    METHOD = 1
    STATIC = 2
    CONSTRUCTOR = 3
    DESTRUCTOR = 4
    GENERIC = 5


def write_generic(writer: Writer, generic: Generic) -> None:
    """
    Generate generic definition and specializations

    Generics are implemented with a macro definition
    which takes in a TYPE argument
    """
    writer.line(f"%define %{generic.name}(TYPE)")
    with writer.indent():
        writer.line(f"%nodefaultctor {generic.name}_##TYPE;")

        # The generic and its specializations are equal in C
        # eg. RzList_int == RzList
        writer.line("%{", f"typedef {generic.name} {generic.name}_##TYPE;", "%}")

        # Treat them differently in SWIG only
        writer.line(f"typedef struct {{}} {generic.name}_##TYPE;")

        writer.line(f"%extend {generic.name}_##TYPE {{")
        with writer.indent():
            for name, method in generic.methods.items():
                write_func(writer, method, name, FuncKind.GENERIC)

            for python_lines in generic.python_methods.values():
                writer.line("%pythoncode %{")
                with writer.indent():
                    writer.line(*python_lines)
                writer.line("%}")
        writer.line("}")
    writer.line("%enddef")

    for specialization in generic.specializations:
        writer.line(f"%{generic.name}({specialization})")

    for specialization, extension in generic.specialization_extensions.items():
        writer.line(f"%extend {generic.name}_{specialization} {{")
        with writer.indent():
            writer.line(*extension)
        writer.line("}")


def stringify_decl(expr: str, ctype: CType, generic: bool = False) -> str:
    """
    Get the spelling of a C declaration for a given name and type
    """

    pointing = False
    while True:
        if isinstance(ctype, CPrimitiveType):
            if ctype.type_.kind == TypeKind.BOOL:
                return f"bool {expr}"
            if generic and ctype.type_.kind == TypeKind.VOID:
                return f"TYPE {expr}"
            return f"{ctype.type_.spelling} {expr}"

        # type_.spelling automatically applies const
        # for other ctypes, do so manually
        if ctype.type_.is_const_qualified():
            expr = "const " + expr

        if isinstance(ctype, CPointerType):
            expr = "*" + expr
            ctype = ctype.pointee
            pointing = True
            continue

        if isinstance(ctype, CRecordType):
            if generic:
                return f"{generic_structs[ctype.decl_spelling].name}_##TYPE {expr}"
            if ctype.generic:
                return f"{ctype.generic.name}_{ctype.specialization} {expr}"
            return f"{class_structs[ctype.decl_spelling].name} {expr}"

        if isinstance(ctype, CTypedefType):
            if isinstance(ctype.canonical, CRecordType) and (
                generic or ctype.canonical.generic
            ):
                ctype = ctype.canonical
                continue
            return f"{ctype.cursor.spelling} {expr}"

        # Wrap in parentheses if the previous node
        # was a pointer (to fix precedence)
        if pointing:
            expr = f"({expr})"
            pointing = False

        if isinstance(ctype, CFunctionType):
            args = ", ".join([stringify_decl("", arg) for arg in ctype.args])
            expr = f"{expr}({args})"
            ctype = ctype.result
        elif isinstance(ctype, CArrayType):
            element_count = ctype.element_count or ""
            expr = f"{expr}[{element_count}]"
            ctype = ctype.element
        else:
            assert_never(ctype)


def write_class(writer: Writer, cls: Class) -> None:
    """
    Generate SWIG class
    """
    writer.line(
        f"typedef struct {cls.struct_name} {cls.name};",
        f"%rename {cls.struct_name} {cls.name};",
    )

    # %rename fields
    for field in cls.fields.values():
        if field.rename:
            writer.line(f"%rename {cls.struct_name}::{field.name} {field.rename};")

    # Main struct
    writer.line(f"struct {cls.struct_name} {{")
    with writer.indent():
        for field in cls.fields.values():
            decl = stringify_decl(field.name, field.ctype)
            writer.line(f"{decl};")
    writer.line("};")

    # un %rename fields
    for field in cls.fields.values():
        if field.rename:
            writer.line(f'%rename {cls.struct_name}::{field.name} "";')

    # Extension
    if len(cls.funcs) != 0 or len(cls.methods) != 0:
        writer.line(f"%extend {cls.struct_name} {{")
        with writer.indent():
            if cls.constructor:
                write_func(
                    writer, cls.constructor, cls.struct_name, FuncKind.CONSTRUCTOR
                )
            if cls.destructor:
                write_func(writer, cls.destructor, cls.struct_name, FuncKind.DESTRUCTOR)
            for name, func in cls.funcs.items():
                write_func(writer, func, name, FuncKind.STATIC)
            for name, method in cls.methods.items():
                write_func(writer, method, name, FuncKind.METHOD)

        writer.line("}")


def write_func(writer: Writer, func: Func, name: str, kind: FuncKind) -> None:
    """
    Generate SWIG function
    """
    # Activate typemaps
    for typemap in func.typemaps:
        typemap_args = ", ".join(f"{arg.type_} {arg.name}" for arg in typemap.args)
        writer.line(f"%{typemap.name}_activate({typemap_args})")

    decl = stringify_decl(
        name,
        func.cfunc.result_ctype,
        isinstance(func, GenericFunc) and func.generic_ret,
    )

    args_outer = []
    if kind in [FuncKind.METHOD, FuncKind.DESTRUCTOR, FuncKind.GENERIC]:
        args_inner = ["$self"]
        args = func.cfunc.args[1:]
    else:
        args_inner = []
        args = func.cfunc.args

    args_nonnull = []  # Used for nullability contract

    for arg in args:
        arg_name = arg.cursor.spelling
        if arg_name == "self":
            arg_name = "_self"

        arg_decl = stringify_decl(
            arg_name,
            arg.ctype,
            isinstance(func, GenericFunc) and arg.cursor.spelling in func.generic_args,
        )

        if arg.default:
            arg_decl += f" = {arg.default}"

        args_outer.append(arg_decl)
        args_inner.append(arg_name)

        if "RZ_NONNULL" in arg.attrs:
            args_nonnull.append(arg_name)

    args_outer_str = ", ".join(args_outer)
    args_inner_str = ", ".join(args_inner)

    # Nullability checking contract
    if args_nonnull:
        writer.line(f"%contract {name}({args_outer_str}) {{", "require:")
        with writer.indent():
            for contract_arg in args_nonnull:
                writer.line(f"{contract_arg} != NULL;")
        writer.line("}")

    if kind in [FuncKind.METHOD, FuncKind.GENERIC]:
        writer.line(f"{decl}({args_outer_str}) {{")
    elif kind == FuncKind.STATIC:
        writer.line(f"static {decl}({args_outer_str}) {{")
    elif kind == FuncKind.CONSTRUCTOR:
        writer.line(f"{name}({args_outer_str}) {{")
    elif kind == FuncKind.DESTRUCTOR:
        writer.line(f"~{name}({args_outer_str}) {{")

    with writer.indent():
        if "RZ_DEPRECATE" in func.cfunc.attrs:
            writer.line(
                f'rizin_try_warn_deprecate("{name}", "{func.cfunc.cursor.spelling}");'
            )

        if kind == FuncKind.GENERIC:
            typecast = stringify_decl(
                "",
                func.cfunc.result_ctype,
                isinstance(func, GenericFunc) and func.generic_ret,
            )
            writer.line(
                f"return ({typecast}){func.cfunc.cursor.spelling}({args_inner_str});"
            )
        else:
            writer.line(f"return {func.cfunc.cursor.spelling}({args_inner_str});")

    writer.line("}")

    # Deactivate typemaps
    for typemap in func.typemaps:
        typemap_args = ", ".join(f"{arg.type_} {arg.name}" for arg in typemap.args)
        writer.line(f"%{typemap.name}_deactivate({typemap_args})")


def write_director(writer: Writer, director: Director) -> None:
    """
    Generate SWIG director
    """
    writer.line(f'%feature("director") {director.name}Director;')

    # Director struct
    writer.line("%inline %{", f"struct {director.name}Director {{")
    with writer.indent():
        for name, func in director.funcs.items():
            decl = stringify_decl(name, func.result_ctype)
            args_outer_str = ", ".join(
                stringify_decl(arg.name, arg.ctype) for arg in func.args
            )
            writer.line(
                f"virtual {decl}({args_outer_str}) {{",
                f'    throw Swig::DirectorPureVirtualException("{name}");',
                "}",
            )

        # Virtual destructor
        writer.line(f"virtual ~{director.name}Director() {{}}")
    writer.line("};", "%}")

    # Global
    writer.line(
        "%{",
        f"static {director.name}Director *SWIG_{director.name}Director = NULL;",
        "%}",
    )

    # Funcs
    writer.line("%{")
    for name, func in director.funcs.items():
        decl = stringify_decl(f"SWIG_{director.name}_{name}", func.result_ctype)
        args_outer_str = ", ".join(
            stringify_decl(arg.name, arg.ctype) for arg in func.args
        )
        args_inner_str = ", ".join(arg.name for arg in func.args)
        writer.line(
            f"{decl}({args_outer_str}) {{",
            f"    return SWIG_{director.name}Director->{name}({args_inner_str});",
            "}",
        )
    writer.line("%}")

    # Builder struct
    writer.line("%inline %{", f"struct {director.name}Builder {{")
    with writer.indent():
        for func_name in director.funcs.keys():
            writer.line(f"bool enable_{func_name};")

        for field_name, field_ctype in director.fields.items():
            decl = stringify_decl(field_name, field_ctype)
            writer.line(f"{decl};")

        writer.line(f"{director.name} *build({director.name}Director *director) {{")
        with writer.indent():
            writer.line(
                f"SWIG_{director.name}Director = director;",
                f"{director.name} *result = ({director.name}*)calloc(1, sizeof({director.name}));",
            )
            for func_name in director.funcs.keys():
                writer.line(
                    f"if (this->enable_{func_name}) {{",
                    f"    result->{func_name} = SWIG_{director.name}_{func_name};",
                    "}",
                )

            for field_name in director.fields.keys():
                writer.line(f"result->{field_name} = this->{field_name};")

            writer.line("return result;")
        writer.line("}")
    writer.line("};", "%}")

    # Python helper
    writer.line("%pythoncode %{", f"def register_{director.name}(director_class):")
    with writer.indent():
        writer.line("fields = [")
        with writer.indent():
            for field_name in director.fields.keys():
                writer.line(f'"{field_name}",')
        writer.line("]")

        writer.line("funcs = [")
        with writer.indent():
            for func_name in director.funcs.keys():
                writer.line(f'"{func_name}",')
        writer.line("]")

        writer.line(
            f"builder = {director.name}Builder()",
            "class_vars = vars(director_class).keys()",
            "for field in fields:",
            "    if field in class_vars:",
            "        attr = getattr(director_class, field)",
            "        setattr(builder, field, attr)",
            "for func in funcs:",
            "    if func in class_vars:",
            '        setattr(builder, f"enable_{func}", True)',
            "director = director_class()",
            "return director, builder.build(director)",
        )
    writer.line("%}")


def write_enum(writer: Writer, enum: Enum) -> None:
    """
    Generate SWIG enum
    """
    writer.line("typedef enum {")
    with writer.indent():
        for name, value in enum.fields.items():
            writer.line(f"{name} = {value},")
    writer.line(f"}} {enum.typedef_name};")



"""
SPDX-FileCopyrightText: 2022 wingdeans <wingdeans@protonmail.com>
SPDX-License-Identifier: LGPL-3.0-only
"""

from typing import List, OrderedDict, DefaultDict, Set, Optional

from functools import cached_property
import os

from clang.cindex import TranslationUnit, Cursor, CursorKind

from cparser_types import CType, wrap_type


### Cursor wrappers ###
class AttrCursor:
    """
    Base class for cursor wrapper with attributes
    """

    cursor: Cursor

    def __init__(self, cursor: Cursor):
        self.cursor = cursor

    @cached_property
    def attrs(self) -> Set[str]:
        """
        Get annotation __attribute__'s on cursor
        """
        return {
            child.spelling
            for child in self.cursor.get_children()
            if child.kind == CursorKind.ANNOTATE_ATTR
        }


class CFuncArg(AttrCursor):
    """
    Function argument cursor wrapper
    """

    default: Optional[str]

    def __init__(self, cursor: Cursor):
        super().__init__(cursor)
        self.default = None

    @cached_property
    def ctype(self) -> CType:
        """
        Get wrapped arg type
        """
        return wrap_type(self.cursor.type)


class CFunc(AttrCursor):
    """
    Function declaration cursor wrapper
    """

    @cached_property
    def args(self) -> List[CFuncArg]:
        """
        Get function args
        """
        return [CFuncArg(arg) for arg in self.cursor.get_arguments()]

    @cached_property
    def result_ctype(self) -> CType:
        """
        Get wrapped result type
        """
        return wrap_type(self.cursor.result_type)


### Configuration ###
rizin_include_path: Optional[str] = None
clang_args: List[str] = []

### Headers ###
headers: List["Header"] = []


class HeaderBuilder:
    """
    Builder class for Header, enabling parallel calls to libclang
    """

    name: str
    filename: str
    extra_filename: Optional[str]

    def __init__(self, name: str):
        self.name = name
        name_segments = name.split("/")

        assert rizin_include_path
        filename = os.path.abspath(os.path.join(rizin_include_path, *name_segments))

        # Fix sdb path if using source librz/include
        if name_segments[0] == "sdb" and not os.path.exists(filename):
            name_segments = ["..", "util", "sdb", "src"] + name_segments[1:]
            filename = os.path.abspath(os.path.join(rizin_include_path, *name_segments))

        assert os.path.exists(filename)
        self.filename = filename

        # Parse ht_inc.h if importing SDB hashtable header
        self.extra_filename = None
        if name_segments[-1] in ["ht_pp.h", "ht_pu.h", "ht_up.h", "ht_uu.h"]:
            self.extra_filename = "ht_inc.h"

    def translation_unit(self) -> TranslationUnit:
        """
        Creates TranslationUnit (for use in the build method)

        TranslationUnit.from_source is parallelizable since ctypes
        releases the GIL, so we leave it to the caller to call this
        method in parallel and later call build sequentially.
        """
        return TranslationUnit.from_source(
            filename=self.filename,
            args=clang_args,
            options=TranslationUnit.PARSE_DETAILED_PROCESSING_RECORD,
        )

    def build(self, translation_unit: TranslationUnit) -> "Header":
        """
        Build the Header instance
        """
        return Header(translation_unit, self)


class Header:
    """
    A parsed C header file
    """

    name: str

    cursors: DefaultDict[CursorKind, OrderedDict[str, Cursor]]
    cursor_kinds: OrderedDict[str, CursorKind]
    cfuncs: OrderedDict[str, CFunc]  # Store __attributes__

    def __init__(self, translation_unit: TranslationUnit, builder: HeaderBuilder):
        headers.append(self)

        self.name = builder.name

        self.cursors = DefaultDict(OrderedDict)
        self.cursor_kinds = OrderedDict()
        self.cfuncs = OrderedDict()

        for cursor in translation_unit.cursor.get_children():
            cursor_file = cursor.location.file
            if not cursor_file:
                continue

            # Skip nodes from other headers
            cursor_file_name = cursor_file.name
            if cursor_file_name != builder.filename and not (
                builder.extra_filename
                and cursor_file_name.endswith(builder.extra_filename)
            ):
                continue

            # Skip `#include` and macro expansion
            if cursor.kind in [
                CursorKind.INCLUSION_DIRECTIVE,
                CursorKind.MACRO_INSTANTIATION,
            ]:
                continue

            # Ignore anonymous declarations
            name = cursor.spelling
            if not name:
                continue

            # Check for redefinitions
            if name in self.cursor_kinds:  # Redefinition
                prev = self.cursors[cursor.kind][name]
                if cursor.kind == CursorKind.STRUCT_DECL:
                    assert prev.kind == CursorKind.STRUCT_DECL
                    assert not any(prev.get_children())  # Should be forward declaration
                elif cursor.kind == CursorKind.MACRO_DEFINITION:
                    assert prev.kind == CursorKind.MACRO_DEFINITION
                    basename = os.path.basename(cursor_file_name)
                    assert (basename, name) in [
                        ("rz_types.h", "__WINDOWS__")
                    ], f"Unexpected redefinition of macro {name} in file {basename}"
                else:
                    raise Exception(
                        f"Unexpected redefinition of symbol: {name}, "
                        f"of kind: {cursor.kind} at {cursor.location}"
                    )

            self.cursor_kinds[name] = cursor.kind
            if cursor.kind == CursorKind.FUNCTION_DECL:
                self.cfuncs[name] = CFunc(cursor)
            else:
                self.cursors[cursor.kind][name] = cursor

    def pop(self, kind: CursorKind, name: str) -> Cursor:
        """
        Remove and return a cursor with the given name and kind
        """
        return self.cursors[kind].pop(name)

    def pop_func(self, name: str) -> CFunc:
        """
        Remove and return a CFunc with the given name and kind
        """
        return self.cfuncs.pop(name)

    def ignore(self, *names: str, prefix: Optional[str] = None) -> None:
        """
        Remove cursors and CFuncs with the given names
        or that have the given prefix
        """
        names_list = list(names)

        if prefix:
            for name in self.cursor_kinds:
                if name.startswith(prefix):
                    names_list.append(name)

        for name in names_list:
            kind = self.cursor_kinds.pop(name)
            if kind == CursorKind.FUNCTION_DECL:
                self.pop_func(name)
            else:
                self.pop(kind, name)
